# w.c.s. - web application for online forms
# Copyright (C) 2005-2010  Entr'ouvert
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301  USA

from qommon import ezt
from cStringIO import StringIO

from quixote import get_request

import qommon.misc
from qommon.storage import StorableObject
from qommon.form import *
from qommon import emails, get_cfg, get_logger
from quixote.directory import Directory
from quixote.html import htmltext
import qommon.errors

from wcs.roles import Role, logged_users_role, get_user_roles
from wcs.formdata import Evolution
from wcs.fields import SubtitleField, TitleField, CommentField, PageField

if not __name__.startswith('wcs.') and not __name__ == "__main__":
    raise ImportError('Import of workflows module must be absolute (import wcs.workflows)')

def lax_int(s):
    try:
        return int(s)
    except (ValueError, TypeError):
        return -1


class TemplatingError(qommon.errors.PublishError):
    def __init__(self, description):
        self.title = _('Templating Error')
        self.description = description

def get_varnames(fields):
    '''Extract variable names for helping people fill their templates.

       Prefer to variable name to the numeric field name.
    '''
    varnames = []
    for field in fields:
        if isinstance(field, (SubtitleField, TitleField, CommentField, PageField)):
                continue
        # add it as f$n$
        label = field.label
        if field.varname:
            varnames.append(('var_%s' % field.varname, label))
        else:
            varnames.append(('f%s' % field.id, label))
    return varnames


class Workflow(StorableObject):
    _names = 'workflows'
    name = None
    details = None
    possible_status = None

    def __init__(self, name = None):
        StorableObject.__init__(self)
        self.name = name
        self.possible_status = []

    def add_status(self, name):
        status = WorkflowStatus(name)

        if self.possible_status:
            status.id = str(max([lax_int(x.id) for x in self.possible_status]) + 1)
        else:
            status.id = '1'
        self.possible_status.append(status)
        return status

    def get_status(self, id):
        for status in self.possible_status:
            if status.id == id:
                return status
        raise KeyError()

    def __setstate__(self, dict):
        self.__dict__.update(dict)
        for s in self.possible_status:
            s.parent = self
            for i, item in enumerate(s.items):
                item.parent = s
                if not item.id:
                    item.id =  '%d' % (i+1)

    def get_endpoint_status(self):
        not_endpoint_status = self.get_not_endpoint_status()
        endpoint_status = [x for x in self.possible_status if x not in not_endpoint_status]
        return endpoint_status

    def get_not_endpoint_status(self):
        not_endpoint_status = []
        for status in self.possible_status:
            endpoint = True
            for item in status.items:
                endpoint = item.endpoint and endpoint
                if endpoint is False:
                    not_endpoint_status.append(status)
                    break
        return not_endpoint_status

    def has_options(self):
        for status in self.possible_status:
            for item in status.items:
                for parameter in item.get_parameters():
                    if not getattr(item, parameter):
                        return True
        return False


class WorkflowStatus:
    id = None
    name = None
    items = None

    def __init__(self, name = None):
        self.name = name
        self.items = []

    def append_item(self, type):
        for klass in item_classes:
            if klass.key == type:
                o = klass()
                if self.items:
                    o.id = str(max([lax_int(x.id) for x in self.items]) + 1)
                else:
                    o.id = '1'
                self.items.append(o)
                break
        else:
            raise KeyError()

    def get_item(self, id):
        for item in self.items:
            if item.id == id:
                return item
        raise KeyError()

    def perform_items(self, formdata):
        url = None
        for item in self.items:
            url = item.perform(formdata) or url
        return url

    def get_action_form(self, filled, user):
        form = Form(enctype='multipart/form-data', use_tokens = False)
        for item in self.items:
            if not item.check_auth(filled, user):
                continue
            item.fill_form(form, filled, user)

        if form.widgets or form.submit_widgets:
            return form
        else:
            return None

    def handle_form(self, form, filled, user):
        #form = self.get_action_form(filled, user)

        evo = Evolution()
        evo.time = time.localtime()
        if user:
            evo.who = user.id
        if not filled.evolution:
            filled.evolution = []

        for item in self.items:
            if hasattr(item, 'by'):
                for role in item.by or []:
                    if role == logged_users_role().id:
                        break
                    if role == '_submitter':
                        if filled.user_id == user.id:
                            break
                        else:
                            continue
                    role = get_role_translation(filled.formdef, role)
                    if role in (user.roles or []):
                        break
                else:
                    continue
            next_url = item.submit_form(form, filled, user, evo)
            if next_url is True:
                break
            if next_url:
                if not form.has_errors():
                    filled.evolution.append(evo)
                    if evo.status:
                        filled.status = evo.status
                    filled.store()
                return next_url

        if form.has_errors():
            return

        filled.evolution.append(evo)

        if evo.status:
            filled.status = evo.status
        filled.store()
        url = filled.perform_workflow()
        if url:
            return url

    def get_subdirectories(self, formdata):
        subdirectories = []
        for item in self.items:
            if item.directory_name:
                subdirectories.append((item.directory_name,
                                       item.directory_class(formdata, item, self)))
        return subdirectories

    def __getstate__(self):
        odict = self.__dict__.copy()
        if odict.has_key('parent'):
            del odict['parent']
        return odict


class WorkflowStatusItem:
    description = 'XX'
    id = None
    endpoint = True
    directory_name = None
    directory_class = None

    def init(cls):
        pass
    init = classmethod(init)

    def render_as_line(self):
        return _(self.description)

    def perform(self, formdata):
        pass

    def fill_form(self, form, formdata, user):
        pass

    def submit_form(self, form, formdata, user, evo):
        pass

    def check_auth(self, formdata, user):
        if not hasattr(self, 'by'):
            return True

        for role in self.by or []:
            if user and role == logged_users_role().id:
                return True
            if not user:
                continue
            if role == '_submitter':
                return (formdata.user_id == user.id)
            role = get_role_translation(formdata.formdef, role)
            if role in (user.roles or []):
                return True

        return False

    def add_parameters_widgets(self, form, parameters, prefix='', formdef=None):
        pass

    def get_parameters(self):
        return ()

    def fill_admin_form(self, form):
        self.add_parameters_widgets(form, self.get_parameters())

    def submit_admin_form(self, form):
        for f in self.get_parameters():
            widget = form.get_widget(f)
            if widget:
                setattr(self, f, widget.parse())

    def compute(self, var):
        if not isinstance(var, basestring):
            return var
        if not var.startswith('='):
            return var
        vars = get_publisher().substitutions.get_context_variables()
        try:
            return eval(var[1:], vars)
        except:
            return var

    def get_substitution_variables(self, formdata):
        return {}

    def __getstate__(self):
        odict = self.__dict__.copy()
        if odict.has_key('parent'):
            del odict['parent']
        return odict


def get_role_translation(formdef, role_name):
    if role_name == '_receiver':
        return formdef.receiver_id
    elif role_name == '_submitter':
        raise Exception('_submitter is not a valid role')
    else:
        return role_name

def render_list_of_roles(roles):
    t = []
    for r in roles:
        if r == logged_users_role().id:
            t.append(logged_users_role().name)
        elif r == '_submitter':
            t.append(_('sender'))
        elif r == '_receiver':
            t.append(_('receiver'))
        else:
            try:
                t.append(Role.get(r).name)
            except KeyError:
                pass
    return ', '.join(t)

item_classes = []
item_types = []

def register_item_class(klass):
    if not klass in item_classes:
        item_classes.append(klass)
        item_types.append((klass.key, klass.description))
        klass.init()


class CommentableWorkflowStatusItem(WorkflowStatusItem):
    description = N_('Commentable')
    key = 'commentable'
    endpoint = False

    by = None

    def render_as_line(self):
        if self.by:
            return _('Commentable by %s') % render_list_of_roles(self.by)
        else:
            return _('Commentable (not completed)')

    def fill_form(self, form, formdata, user):
        if not 'comment' in [x.name for x in form.widgets]:
            form.add(TextWidget, 'comment', title = _('Comment'), required = False,
                    cols = 40, rows = 10)
            form.get_widget('comment').attrs['class'] = 'comment'
            form.add_submit(self.id, _('Add Comment'))

    def submit_form(self, form, formdata, user, evo):
        if form.get_widget('comment'):
            evo.comment = form.get_widget('comment').parse()

    def fill_admin_form(self, form):
        if self.by and not type(self.by) is list:
            self.by = None
        self.add_parameters_widgets(form, self.get_parameters())

    def get_parameters(self):
        return ('by',)

    def add_parameters_widgets(self, form, parameters, prefix='', formdef=None):
        if 'by' in parameters:
            form.add(WidgetList, '%sby' % prefix, title = _('By'),
                element_type=SingleSelectWidget,
                value=self.by,
                add_element_label=_('Add Role'),
                element_kwargs={'render_br': False,
                'options': [(None, '---'),
                            ('_submitter', _('Sender')),
                            ('_receiver', _('Receiver')),
                            (logged_users_role().id, logged_users_role().name),
                            (None, '----')] + get_user_roles()})

register_item_class(CommentableWorkflowStatusItem)


class ChoiceWorkflowStatusItem(WorkflowStatusItem):
    description = N_('Choice')
    key = 'choice'
    endpoint = False

    label = None
    status = None
    by = None

    def render_as_line(self):
        if self.label:
            if self.by:
                return _('Choice "%(label)s" by %(by)s') % \
                        { 'label' : self.label, 
                          'by' : render_list_of_roles(self.by) }
            else:
                return _('Choice "%s"') % self.label
        else:
            return _('Choice (not completed)')

    def fill_form(self, form, formdata, user):
        form.add_submit(self.id, self.label)

    def submit_form(self, form, formdata, user, evo):
        if form.get_submit() == self.id:
            wf_status = [x for x in self.parent.parent.possible_status \
                         if str(x.id) == str(self.status)][0]
            evo.status = 'wf-%s' % wf_status.id
            form.clear_errors()
            return True # get out of processing loop

    def add_parameters_widgets(self, form, parameters, prefix='', formdef=None):
        if 'label' in parameters:
            form.add(StringWidget, '%slabel' % prefix, title = _('Label'), value = self.label)
        if 'status' in parameters:
            form.add(SingleSelectWidget, '%sstatus' % prefix, title = _('Status'), value = self.status,
                options = [(None, '---')] + [(x.id, x.name) for x in self.parent.parent.possible_status])
        if 'by' in parameters:
            form.add(WidgetList, '%sby' % prefix, title = _('By'), element_type = SingleSelectWidget,
                value = self.by,
                add_element_label = _('Add Role'),
                element_kwargs = {'render_br': False,
                'options': [(None, '---'),
                            ('_submitter', _('Sender')),
                            ('_receiver', _('Receiver')),
                            (logged_users_role().id, logged_users_role().name),
                            (None, '----')] + get_user_roles()})

    def get_parameters(self):
        return ('by', 'status', 'label')

register_item_class(ChoiceWorkflowStatusItem)


class JumpOnSubmitWorkflowStatusItem(WorkflowStatusItem):
    description = N_('Jump on Submit')
    key = 'jumponsubmit'

    status = None

    def submit_form(self, form, formdata, user, evo):
        if form.is_submitted() and not form.has_errors():
            wf_status = [x for x in self.parent.parent.possible_status if x.id == self.status][0]
            evo.status = 'wf-%s' % wf_status.id

    def add_parameters_widgets(self, form, parameters, prefix='', formdef=None):
        if 'status' in parameters:
            form.add(SingleSelectWidget, '%sstatus' % prefix, title = _('Status'), value = self.status,
                options = [(None, '---')] + [(x.id, x.name) for x in self.parent.parent.possible_status])

    def get_parameters(self):
        return ('status',)
register_item_class(JumpOnSubmitWorkflowStatusItem)


class SendmailWorkflowStatusItem(WorkflowStatusItem):
    description = N_('Send mail')
    key = 'sendmail'

    to = None
    subject = None
    body = None

    comment = None

    def render_as_line(self):
        if self.to:
            return _('Send mail to %s') % render_list_of_roles(self.to)
        else:
            return _('Send mail (not completed)')

    def get_parameters(self):
        return ('to', 'subject', 'body')

    def fill_admin_form(self, form):
        self.add_parameters_widgets(form, self.get_parameters())

    def add_parameters_widgets(self, form, parameters, prefix='', formdef=None):
        if 'to' in parameters:
            form.add(WidgetList, '%sto' % prefix, title=_('To'),
                     element_type=SingleSelectWidget,
                     value=self.to,
                     add_element_label=_('Add Role'),
                     element_kwargs={'render_br': False,
                     'options': [(None, '---'),
                                 ('_submitter', _('Sender')),
                                 ('_receiver', _('Receiver')), ] + get_user_roles()})
        if 'subject' in parameters:
            form.add(StringWidget, '%ssubject' % prefix, title=_('Subject'),
                     value=self.subject, size=40)
        if 'body' in parameters:
            form.add(TextWidget, '%sbody' % prefix, title=_('Body'),
                     value=self.body, cols=80, rows=10,
                     hint=_('Available variables: url, url_status, details, name, number, comment, field_NAME'))


    def perform(self, formdata):
        if not self.to:
            return
        if not self.subject:
            return

        url = formdata.get_url()
        try:
            mail_body = template_on_formdata(formdata, self.compute(self.body))
        except ezt.EZTException:
            get_logger().error('error in template for email body [%s], mail could not be generated' % url)
            return

        try:
            mail_subject = template_on_formdata(formdata, self.compute(self.subject))
        except ezt.EZTException:
            get_logger().error('error in template for email subject [%s], mail could not be generated' % url)
            return

        users_cfg = get_cfg('users', {})

        # this works around the fact that parametric workflows only support
        # string values, so if we get set a string, we convert it here to an
        # array.
        if isinstance(self.to, basestring):
            self.to = [self.to]

        addresses = []
        for dest in self.to:
            dest = self.compute(dest)

            if '@' in str(dest):
                addresses.append(dest)
                continue

            if dest == '_submitter':
                field_email = users_cfg.get('field_email') or 'email'
                done = False
                if formdata.user:
                    if formdata.user.email:
                        addresses.append(formdata.user.email)
                        done = True
                    elif formdata.user.form_data and formdata.user.form_data.get(field_email):
                        addresses.append(formdata.user.form_data.get(field_email))
                        done = True

                if not done:
                    # if there is no user, or user has no email address, look
                    # up in submitted form for one that would hold the user
                    # email (the one set to be prefilled by user email)
                    fields = formdata.formdef.fields
                    for field in fields:
                        if not hasattr(field, 'prefill'):
                            continue
                        if field.prefill and field.prefill.get('type') == 'user':
                            if field.prefill.get('value') == field_email:
                                v = formdata.data.get(field.id)
                                if v:
                                    addresses.append(v)
                                    done = True
                                break

                continue

            dest = get_role_translation(formdata.formdef, dest)

            try:
                role = Role.get(dest)
            except KeyError:
                continue
            addresses.extend(role.get_emails())

        if not addresses:
            return

        if len(addresses) > 1:
            emails.email(mail_subject, mail_body, email_rcpt=None,
                    bcc=addresses,
                    exclude_current_user=False,
                    fire_and_forget=True)
        else:
            emails.email(mail_subject, mail_body, email_rcpt=addresses,
                    exclude_current_user=False,
                    fire_and_forget=True)
register_item_class(SendmailWorkflowStatusItem)


def template_on_formdata(formdata, template, process=None):
    dict = {}
    dict.update(get_publisher().substitutions.get_context_variables())
    dict['url'] = formdata.get_url()
    dict['url_status'] = '%sstatus' % formdata.get_url()
    dict['details'] = formdata.formdef.get_detailed_email_form(formdata, dict['url'])
    dict['name'] = formdata.formdef.name
    dict['number'] = formdata.id
    if formdata.evolution and formdata.evolution[-1].comment:
        dict['comment'] = formdata.evolution[-1].comment
    else:
        dict['comment'] = ''
    dict.update(formdata.get_as_dict())

    if process:
        for key in dict:
            dict[key] = process(dict[key])

    processor = ezt.Template(compress_whitespace=False)
    processor.parse(template or '')

    fd = StringIO()
    processor.generate(fd, dict)

    return fd.getvalue()

class SendSMSWorkflowStatusItem(WorkflowStatusItem):
    description = N_('Send SMS')
    key = 'sendsms'

    to = None
    body = None

    def render_as_line(self):
        return _('Send SMS')

    def fill_admin_form(self, form):
        self.add_parameters_widgets(form, self.get_parameters())

    def get_parameters(self):
        return ('to', 'body')

    def add_parameters_widgets(self, form, parameters, prefix='', formdef=None):
        if 'to' in parameters:
            form.add(WidgetList, '%sto' % prefix, title=_('To'), element_type=StringWidget,
                     value=self.to, add_element_label=_('Add Number'),
                     element_kwargs = {'render_br': False})
        if 'body' in parameters:
            form.add(TextWidget, '%sbody' % prefix, title=_('Body'), value=self.body, cols=80, rows=10,
                     hint = _('Available variables: url, url_status, details, name, number, comment, field_NAME'))

    def perform(self, formdata):
        if not self.to:
            return
        if not self.body:
            return

        try:
            sms_body = template_on_formdata(formdata, self.compute(self.body))
        except ezt.EZTException:
            url = formdata.get_url()
            get_logger().error('error in template for sms [%s], sms could not be generated' % url)
            return

        from qommon.sms import SMS

        sms_cfg = get_cfg('sms', {})
        sender = sms_cfg.get('sender', 'AuQuotidien')[:11]
        try:
            SMS().send([self.compute(x) for x in self.to], sms_body[:160], sender)
        except errors.SMSError, e:
            get_logger().error(e)

register_item_class(SendSMSWorkflowStatusItem)


class DisplayMessageWorkflowStatusItem(WorkflowStatusItem):
    description = N_('Display message')
    key = 'displaymsg'

    message = None

    def get_message(self, filled):
        if not self.message:
            return ''
        tmpl = ezt.Template()
        tmpl.parse(self.message)

        dict = {}
        dict.update(get_publisher().substitutions.get_context_variables())
        dict['date'] = misc.localstrftime(filled.receipt_time)
        dict['number'] = filled.id
        if filled.formdef.receiver and filled.formdef.receiver.details:
            dict['receiver'] = filled.formdef.receiver.details.replace('\n', '<br />')

        fd = StringIO()
        tmpl.generate(fd, dict)
        msg = fd.getvalue()

        return msg

    def add_parameters_widgets(self, form, parameters, prefix='', formdef=None):
        if 'message' in parameters:
            form.add(TextWidget, '%smessage' % prefix, title = _('Message'),
                value = self.message, cols = 80, rows = 10)

    def get_parameters(self):
        return ('message',)

register_item_class(DisplayMessageWorkflowStatusItem)


class RedirectToStatusWorkflowStatusItem(WorkflowStatusItem):
    description = N_('Redirect to Status Page')
    key = 'redirectstatus'

    backoffice = False

    def perform(self, formdata):
        if not get_request().user:
            return None
        return formdata.get_url(self.backoffice)

    def add_parameters_widgets(self, form, parameters, prefix='', formdef=None):
        if 'backoffice' in parameters:
            form.add(CheckboxWidget, '%sbackoffice' % prefix,
                title = _('Redirect to backoffice page'),
                value = self.backoffice)

    def get_parameters(self):
        return ('backoffice',)

register_item_class(RedirectToStatusWorkflowStatusItem)


class EditableWorkflowStatusItem(WorkflowStatusItem):
    description = N_('Editable')
    key = 'editable'
    endpoint = False

    by = None
    status = None
    label = None

    def render_as_line(self):
        if self.by:
            return _('Editable by %s') % render_list_of_roles(self.by)
        else:
            return _('Editable (not completed)')

    def fill_form(self, form, formdata, user):
        label = self.label
        if not label:
            label = _('Edit Form')
        form.add_submit(self.id, label)

    def submit_form(self, form, formdata, user, evo):
        if form.get_submit() == self.id:
            return formdata.get_url() + 'wfedit'

    def add_parameters_widgets(self, form, parameters, prefix='', formdef=None):
        if 'by' in parameters:
            form.add(WidgetList, '%sby' % prefix, title = _('By'), element_type = SingleSelectWidget,
                value = self.by,
                add_element_label = _('Add Role'),
                element_kwargs = {'render_br': False,
                'options': [(None, '---'),
                            ('_submitter', _('Sender')),
                            ('_receiver', _('Receiver')),
                            (logged_users_role().id, logged_users_role().name),
                            (None, '----')] + get_user_roles()})
        if 'status' in parameters:
            # XXX: look into this one, as None is a perfectly valid value, and
            # it would put this question in the 'workflow options' part.
            form.add(SingleSelectWidget, '%sstatus' % prefix, title = _('Status After Edit'), value = self.status,
                hint = _("Don't select any if you don't want status change processing"),
                options = [(None, '---')] + [(x.id, x.name) for x in self.parent.parent.possible_status])
        if 'label' in parameters:
            form.add(StringWidget, '%slabel' % prefix, title = _('Button Label'), value = self.label)

    def get_parameters(self):
        return ('by', 'status', 'label')

register_item_class(EditableWorkflowStatusItem)

class ExportToModelDirectory(Directory):
    _q_exports = [ '' ]

    def __init__(self, formdata, wfstatusitem, wfstatus):
        self.formdata = formdata
        self.wfstatusitem = wfstatusitem

    def _q_index(self):
        if not self.wfstatusitem.model_file:
            raise TemplatingError(_('No model defined for this action'))
        response = get_response()
        response.content_type = self.wfstatusitem.model_file.content_type
        response.set_header('location', '..')

        if response.content_type != 'text/html':
            response.set_header('content-disposition',
                    'attachment; filename="%s"' %
                        self.wfstatusitem.model_file.base_filename)
        return self.wfstatusitem.apply_template_to_formdata(self.formdata)

def char2rtf(c):
    if ord(c) < 128:
        return c
    else:
        return '\u%d?' % ord(c)

def str2rtf(s):
    s = ''.join([ char2rtf(c) for c in s])
    return '{\uc1{%s}}' % s

def rtf_process(value):
    if value is None:
        return None
    return str2rtf(unicode(str(value), get_publisher().site_charset))

class ExportToModel(WorkflowStatusItem):
    description = N_('Export to model')
    key = 'export_to_model'

    label = None
    model_file = None
    directory_class = ExportToModelDirectory

    def render_as_line(self):
        if self.label:
            if self.model_file:
                model = _('with model named %(file_name)s of %(size)s bytes') % {
                    'file_name': self.model_file.base_filename,
                    'size': self.model_file.size }
            else:
                model = _('no model set')
            return _('Export to model, labeled %(label)s, %(model)s') % {
                    'label': self.label,
                    'model': model }
        else:
            return _('Export to model (not completed)')

    def fill_form(self, form, formdata, user):
        label = self.label
        if not label:
            label = _('Export to model')
        form.add_submit(self.id, label)

    def submit_form(self, form, formdata, user, evo):
        if form.get_submit() == self.id:
            if not evo.comment:
                evo.comment = _('Form exported in a model')
            return formdata.get_url() + self.get_directory_name()

    def model_file_validation(self, upload):
        if upload.content_type and upload.content_type == 'application/rtf':
            return True, ''
        if (upload.content_type and upload.content_type == 'application/octet-stream') or \
                upload.content_type is None:
            if upload.base_filename and upload.base_filename.endswith('.rtf'):
                return True, ''
        return False, _('Only RTF files can be used')

    def add_parameters_widgets(self, form, parameters, prefix='', formdef=None):
        if 'label' in parameters:
            form.add(StringWidget, '%slabel' % prefix, title = _('Button Label'), value = self.label)
        if 'model_file' in parameters:
            ids = (self.parent.parent.id, self.parent.id, self.id)
            if formdef:
                hint = htmltext('%s: <ul class="varnames">') % _('Available variables')
                varnames = get_varnames(formdef.fields)
                for pair in varnames:
                    hint += htmltext('<li><tt class="varname">[%s]</tt> <label>%s</label></span></li>') % pair
                hint += htmltext('</ul>')
                ids = (formdef.id,) + ids
                filename = 'export_to_model-%s-%s-%s-%s.upload' % ids
            else:
                hint = _('You can use variables in your model using '
                        'the [variable] syntax, available variables depends on the form.')
                filename = 'export_to_model-%s-%s-%s.upload' % ids
            widget_name = '%smodel_file' % prefix
            if formdef and formdef.workflow_options and \
                    formdef.workflow_options.get(widget_name) is not None:
                value = formdef.workflow_options.get(widget_name)
            else:
                value = self.model_file
            if value:
                hint = htmltext('<div>%s: <a href="?file=%s">%s</a></div>') % \
                        (_('Current value'), widget_name, value.base_filename) + hint
            form.add(UploadWidget, widget_name, directory='models',
                    filename=filename, title=_('Model'), hint=hint,
                    validation=self.model_file_validation, value=value)

    def get_directory_name(self):
        return qommon.misc.simplify(self.label or 'export_to_model', space='_')
    directory_name = property(get_directory_name)

    def apply_template_to_formdata(self, formdata):
        process = None
        if self.model_file.base_filename.endswith('.rtf'):
            process = rtf_process
        try:
            return template_on_formdata(formdata, self.model_file.get_file().read(),
                    process=process)
        except ezt.UnknownReference, e:
            url = formdata.get_url()
            get_logger().error('error in template for export to model [%s], unknown reference %s' % (url, str(e)))
            raise TemplatingError(_('Error in the template, reference %s is unknown') % str(e))
        except ezt.EZTException, e:
            url = formdata.get_url()
            get_logger().error('error in template for export to model [%s], model could not be generated' % url)
            raise TemplatingError(_('Unknown error in the template: %s') % str(e))

    def get_parameters(self):
        return ('label', 'model_file')

register_item_class(ExportToModel)


def load_extra():
    import wf.aggregation_email
    import wf.timeout_jump
    import wf.anonymous_access
